#include <QDebug>
#include <ClassDesc.hpp>

ClassDesc::ClassDesc(unsigned long startItem,
					 QString const& spec,
					 QString const& name)
:	startItem_(startItem),
	spec_(spec),
	name_(name),
	nameInx_(startItem),
	openInx_(0),
	closeInx_(0)
{
}

ClassDesc::~ClassDesc()
{
	// do not delete the nested class descriptors: assume that
	// the caller takes care of this!
}

ClassDesc::Ptr ClassDesc::create(unsigned long& i,
								 ClassItem::Vec const& items,
								 QString const& baseName,
								 QStringList& incomplete)
{
	ClassDesc::Ptr cl;
	bool bracketOpened = false,
		 bracketClosed = false;

	qDebug() << QString("Calling create(%1, <%2>)").arg(i).arg(baseName);

	for(; (i < (unsigned long)items.size()) &&
		  (!cl || !bracketClosed);
		++i)
	{
		ClassItem const& item = items.at(i);
		QString shortKey = item.first;
		shortKey.remove(QRegExp("\\s"));

		qDebug() << QString("loop: %1 - <%2>").arg(i).arg(item.first);

		if("/" == item.first.left(1))
		{
			// ignore comment
			qDebug() << "comment";

			continue;
		}

		else if(("class" == item.first) || ("struct" == item.first))
		{
			qDebug() << "class/struct";
			
			// this is only a forwarder -> ignore
			if(item.second.trimmed().endsWith(';'))
				continue;

			// if we have no class declaration yet, we create one
			if(!cl)
			{
				cl = ClassDesc::Ptr(new ClassDesc(i,
						item.first,
						QString("%1%2%3")
						.arg(baseName)
						.arg(baseName.isEmpty() ? "" : "::")
						.arg(item.name())));
				continue;
			}

			// if we have a class declaration already, but no
			// opening bracket yet, this is a syntax error
			// (incomplete class declaration)
			// the new item will start a new descriptor
			if(!bracketOpened)
			{
				ClassItem const& item = items.at(cl->startItem_);
				incomplete.append(QString("%1 %2").arg(item.key()).arg(item.name()));
				cl = ClassDesc::Ptr(new ClassDesc(i,
						item.first,
						QString("%1%2%3")
						.arg(baseName)
						.arg(baseName.isEmpty() ? "" : "::")
						.arg(item.name())));
				continue;
			}

			// if we are getting here, this is the beggining of
			// a nested class declaration (or several!)
			// -> get them iteratively
			// note that we do not care whether we are in a public,
			// private or whatever segment here as long as we are
			// between the opening and closing class declaration brackets
			while(true)
			{
				ClassDesc::Ptr clNested = create(i, items, cl->name_, incomplete);
				if(!clNested)
					break;
				cl->classes_.push_back(clNested);
				QString nm;
				if(i < (unsigned long)items.size())
				{
					ClassItem const& item = items.at(i);
					nm = item.first;
				}
				if(("class" != nm) && ("struct" != nm))
				{
					--i;	// compensate for the ++i that will follow in the loop
					break;
				}
			}
		}

		else if(!cl)
		{
			qDebug() << "0 == cl";
			
			// no class yet -> just go on
			continue;
		}

		else if("{" == item.first)
		{
			qDebug() << "{";
			
			// this will really start a declaration
			bracketOpened = true;
			cl->openInx_ = i;
		}

		else if(!bracketOpened)
		{
			qDebug() << "!bracketOpened";
			
			// if we have a new class and the first thing that
			// follows is not an opening bracket, this declaration is
			// not correct -> drop it as "incomplete"
			ClassItem const& item = items.at(cl->startItem_);
			incomplete.append(QString("%1 %2").arg(item.key()).arg(item.name()));
			cl = ClassDesc::Ptr();
			--i;	// redo this item at the next upper level, if any
			return ClassDesc::Ptr();
		}

		else if("public:" == shortKey)
		{
			qDebug() << "public";
			
			// starting public segment
			cl->publicSegs_.push_back(i);
		}

		else if("protected:" == shortKey)
		{
			qDebug() << "protected";

			// starting protected segment
			cl->protectedSegs_.push_back(i);
		}

		else if("private:" == shortKey)
		{
			qDebug() << "private";

			// starting private segment
			cl->privateSegs_.push_back(i);
		}

		else if("publicslots:" == shortKey)
		{
			qDebug() << "public slots";
			
			// starting public slots segment
			cl->publicSlotsSegs_.push_back(i);
		}

		else if("protectedslots:" == shortKey)
		{
			qDebug() << "protected slots";

			// starting protected slots segment
			cl->protectedSlotsSegs_.push_back(i);
		}

		else if("privateslots:" == shortKey)
		{
			qDebug() << "private slots";

			// starting private slots segment
			cl->privateSlotsSegs_.push_back(i);
		}

		else if("signals:" == shortKey)
		{
			qDebug() << "signals";

			// starting signals segment
			cl->signalsSegs_.push_back(i);
		}

		else if('}' == item.first.at(0))
		{
			qDebug() << "}";

			// finishing class declaration
			bracketClosed = true;
			cl->closeInx_ = i;
		}
	}

	// if we finished without finishing the class declaration,
	// add it to the incompletes and delete the descriptor
	if(cl && !bracketClosed)
	{
		ClassItem const& item = items.at(cl->startItem_);
		incomplete.append(QString("%1 %2").arg(item.key()).arg(item.name()));
		qDebug() << QString("incomplete class: <%1>").arg(incomplete.back());
		cl = ClassDesc::Ptr();
	}

	qDebug() << QString("returning cl = %1").arg((unsigned long)cl.data());

	// whatever we have: return it now
	return cl;
}

ClassDesc::Vec ClassDesc::classVec() const
{
	Vec cv;
	for(ConstIt it = classes_.begin(); it != classes_.end(); ++it)
	{
		cv.append(*it);
		cv += (*it)->classVec();
	}
	return cv;
}

QStringList ClassDesc::segs() const
{
	QStringList sgs;
	sgs.append(QString("name: %1").arg(nameInx_));
	sgs.append(QString("open: %1").arg(openInx_));
	sgs.append(QString("close: %1").arg(closeInx_));
	sgs.append(QString("publicSegs: %1").arg(inxList(publicSegs_)));
	sgs.append(QString("protectedSegs: %1").arg(inxList(protectedSegs_)));
	sgs.append(QString("privateSegs: %1").arg(inxList(privateSegs_)));
	sgs.append(QString("publicSlotsSegs: %1").arg(inxList(publicSlotsSegs_)));
	sgs.append(QString("protectedSlotsSegs: %1").arg(inxList(protectedSlotsSegs_)));
	sgs.append(QString("privateSlotsSegs: %1").arg(inxList(privateSlotsSegs_)));
	sgs.append(QString("signalsSegs: %1").arg(inxList(signalsSegs_)));
	return sgs;
}

QString ClassDesc::inxList(InxVec const& iv)
{
	QStringList il;
	for(InxVec::const_iterator it = iv.begin(); it != iv.end(); ++it)
		il.append(QString("%1").arg(*it));
	return il.join(" ");
}
