#include <QRegExp>
#include <QPalette>
#include <QFileInfo>
#include <QFileDialog>
#include <DlgFunctionProperties.hpp>

DlgFunctionProperties::DlgFunctionProperties(QStringList const& clist,
											 QString const& sourceFile,
											 QWidget* parent)
:	QDialog(parent)
{
	ui.setupUi(this);

	for(QStringList::const_iterator it = clist.begin(); it != clist.end(); ++it)
		ui.comboBoxClass->addItem(*it);

	ui.comboBoxAccessSpec->addItem("public");
	ui.comboBoxAccessSpec->addItem("protected");
	ui.comboBoxAccessSpec->addItem("private");
	ui.comboBoxAccessSpec->addItem("public slots");
	ui.comboBoxAccessSpec->addItem("protected slots");
	ui.comboBoxAccessSpec->addItem("private slots");
	ui.comboBoxAccessSpec->addItem("signals");
	ui.comboBoxAccessSpec->setCurrentIndex(0);

	ui.lineEditFunctionName->setText("newFunction");
	ui.lineEditReturnType->setText("void");

	connect(ui.comboBoxClass, SIGNAL(currentIndexChanged(int)),
			this, SLOT(updateEnableControlsClass(int)));
	connect(ui.checkBoxConst, SIGNAL(stateChanged(int)),
			this, SLOT(updateEnableControlsConst(int)));
	connect(ui.checkBoxStatic, SIGNAL(stateChanged(int)),
			this, SLOT(updateEnableControlsStatic(int)));
	connect(ui.checkBoxInline, SIGNAL(stateChanged(int)),
			this, SLOT(updateEnableControlsInline(int)));
	connect(ui.lineEditSourceFile, SIGNAL(textChanged(QString const&)),
			this, SLOT(updateSourceFileInfo(QString const&)));
	connect(ui.pushButtonBrowse, SIGNAL(released()),
			this, SLOT(browseSource()));
	
	ui.lineEditSourceFile->setText(sourceFile);
	if(sourceFile.isEmpty())
		ui.lineEditSourceFile->setEnabled(false);
}

void DlgFunctionProperties::updateEnableControlsClass(int sel)
{
	bool global = (ui.comboBoxClass->count() - 1) == sel;
	ui.checkBoxConst->setEnabled(!global);
	ui.checkBoxVirtual->setEnabled(!global);
	ui.comboBoxAccessSpec->setEnabled(!global);
}

void DlgFunctionProperties::updateEnableControlsConst(int state)
{
	ui.checkBoxStatic->setEnabled(Qt::Checked != state);
}

void DlgFunctionProperties::updateEnableControlsStatic(int state)
{
	ui.checkBoxConst->setEnabled(Qt::Checked != state);
}

void DlgFunctionProperties::updateEnableControlsInline(int state)
{
	ui.lineEditSourceFile->setEnabled(Qt::Checked != state);
	ui.labelSourceFileInfo->setEnabled(Qt::Checked != state);
	ui.pushButtonBrowse->setEnabled(Qt::Checked != state);
}

void DlgFunctionProperties::updateSourceFileInfo(QString const& text)
{
	QFileInfo fi(text);
	QString msg;
	QColor col;

	if(text.isEmpty())
	{
		msg = QObject::tr("No source file defined");
		col = Qt::red;
	}
	else if(!fi.dir().exists())
	{
		msg = QObject::tr("Source directory does not exist - will be created");
		col = Qt::red;
	}
	else if(!fi.exists())
	{
		msg = QObject::tr("Source file does not exist - will be created");
		col = Qt::red;
	}
	else
	{
		msg = QObject::tr("Source file exists");
		col = Qt::black;
	}

	QPalette pal = ui.labelSourceFileInfo->palette();
	pal.setColor(QPalette::WindowText, col);
	ui.labelSourceFileInfo->setPalette(pal);
	ui.labelSourceFileInfo->setText(msg);
}

void DlgFunctionProperties::browseSource()
{
	QFileDialog* fdlg = new QFileDialog(this, QObject::tr("Select source file"),
										ui.lineEditSourceFile->text());
	if(QDialog::Accepted != fdlg->exec())
		return;

	QStringList fs = fdlg->selectedFiles();
	if(fs.isEmpty())
		return;

	QString fn = fs.front();
	ui.lineEditSourceFile->setText(QString("%1\\%2")
								   .arg(fdlg->directory().canonicalPath())
								   .arg(fn));
}

QString DlgFunctionProperties::accessSpec() const
{
	if(ui.comboBoxAccessSpec->isEnabled())
		return ui.comboBoxAccessSpec->currentText();
	else
		return "";
}

QString DlgFunctionProperties::functionName() const
{
	return ui.lineEditFunctionName->text();
}

QString DlgFunctionProperties::parameters() const
{
	return ui.lineEditParameters->text();
}

QString DlgFunctionProperties::parametersWithoutDefaults() const
{
	QString pars = parameters();
	QRegExp rex("\\s*=\\s*.+,?");
	rex.setMinimal(true);
	pars.replace(rex, "");
	return pars;
}

QString DlgFunctionProperties::returnType() const
{
	if(ui.lineEditReturnType->text().isEmpty())
		return QString("void");
	else
		return ui.lineEditReturnType->text();
}

QString DlgFunctionProperties::className() const
{
	QString cl = ui.comboBoxClass->currentText();
	if((0 == cl.indexOf("class")) || (0 == cl.indexOf("struct")))
		cl = cl.mid(cl.indexOf(" ") + 1);
	return cl;
}

bool DlgFunctionProperties::isGlobal() const
{
	return ui.comboBoxClass->itemText(ui.comboBoxClass->count() - 1) == className();
}

bool DlgFunctionProperties::isConst() const
{
	return ui.checkBoxConst->isEnabled() &&
		   ui.checkBoxConst->isChecked();
}

bool DlgFunctionProperties::isVirtual() const
{
	return ui.checkBoxVirtual->isEnabled() &&
		   ui.checkBoxVirtual->isChecked();
}

bool DlgFunctionProperties::isStatic() const
{
	return ui.checkBoxStatic->isEnabled() &&
		   ui.checkBoxStatic->isChecked();
}

bool DlgFunctionProperties::isInline() const
{
	return ui.checkBoxInline->isEnabled() &&
		   ui.checkBoxInline->isChecked();
}

QString DlgFunctionProperties::sourceFile() const
{
	if(isGlobal())
		return "";
	return ui.lineEditSourceFile->text();
}
