#include <QDebug>
#include <QPair>
#include <QVector>
#include <QColorDialog>
#include <QToolBar>
#include <QFileInfo>
#include <QDir>
#include <QRegExp>
#include <QTextStream>
#include <QMessageBox>
#include <ClassItem.hpp>
#include <ClassDesc.hpp>
#include <juffed/PluginSettings.h>
#include <DlgClassProperties.hpp>
#include <DlgFunctionProperties.hpp>
#include <DlgCppSettings.hpp>
#include <DateFormatDlg.h>
#include <TextModPlugin.h>

//#define DEBUG_LOG "D:\\C\\Logs\\TextModPlugin.log"

#ifdef DEBUG_LOG
void myMessageOutput(QtMsgType type, const char* msg)
{
	QFile df(DEBUG_LOG);
	if(!df.open(QIODevice::ReadWrite | QIODevice::Text | QIODevice::Append))
		return;
	df.write(msg);
	df.write("\n");
	df.close();
}
#endif

TextModPlugin::TextModPlugin()
:	ctxMenuCpp_(0),
	ctxMenuHtml_(0),
	menu_(0),
	toolBar_(0),
	dlgCppSettings_(0)
{
	//	context menu
	ctxMenuCpp_ = new QMenu(tr("C++"));
	ctxMenuCpp_->addAction(tr("New Class"), this, SLOT(cppNewClass()));
	ctxMenuCpp_->addAction(tr("New Function"), this, SLOT(cppNewFunc()));
	ctxMenuHtml_ = new QMenu(tr("Selected text..."));
	ctxMenuHtml_->addAction(tr("TO UPPER CASE"), this, SLOT(toUpper()));
	ctxMenuHtml_->addAction(tr("to lower case"), this, SLOT(toLower()));
	QMenu* htmlMenu = ctxMenuHtml_->addMenu("HTML");
	htmlMenu->addAction(tr("<a href ... >"), this, SLOT(htmlHref()));
	htmlMenu->addAction(tr("<b>...</b>"), this, SLOT(htmlBold()));
	htmlMenu->addAction(tr("<i>...</i>"), this, SLOT(htmlItalic()));
	htmlMenu->addAction(tr("<u>...</u>"), this, SLOT(htmlUnderline()));
	htmlMenu->addAction(tr("<div>...</div>"), this, SLOT(htmlDiv()));
	htmlMenu->addAction(tr("<ul>...</ul>"), this, SLOT(htmlUList()));
	htmlMenu->addAction(tr("<ol>...</ol>"), this, SLOT(htmlOList()));

	//	main menu
	menu_ = new QMenu(tr("Insert"));
	menu_->addAction(tr("Current Date and Time"), this, SLOT(date()));
	QMenu* cppMenu = menu_->addMenu("C++");
	cppMenu->addAction(tr("New Class"), this, SLOT(cppNewClass()));
	cppMenu->addAction(tr("New Function"), this, SLOT(cppNewFunc()));
	htmlMenu = menu_->addMenu("HTML");
	htmlMenu->addAction(tr("File template"), this, SLOT(htmlTemplate()));
	htmlMenu->addAction(tr("<table> ... "), this, SLOT(htmlTable()));
	htmlMenu->addAction(tr("<a href ... >"), this, SLOT(htmlHref()));
	htmlMenu->addAction(tr("<img ... >"), this, SLOT(htmlImg()));
	htmlMenu->addAction(tr("<b>...</b>"), this, SLOT(htmlBold()));
	htmlMenu->addAction(tr("<i>...</i>"), this, SLOT(htmlItalic()));
	htmlMenu->addAction(tr("<u>...</u>"), this, SLOT(htmlUnderline()));
	htmlMenu->addAction(tr("<div>...</div>"), this, SLOT(htmlDiv()));
	htmlMenu->addAction(tr("<ul>...</ul>"), this, SLOT(htmlUList()));
	htmlMenu->addAction(tr("<ol>...</ol>"), this, SLOT(htmlOList()));

	// toolbar
	toolBar_ = new QToolBar("TextMod");
	toolBar_->addAction(QIcon(":color_icon"), "Insert a color", this, SLOT(callColorDlg()));

	// cpp settings
	sett_.load(this);
	
	dlgCppSettings_ = new DlgCppSettings(sett_);

	// debug
#ifdef DEBUG_LOG
	QFile::remove(DEBUG_LOG);
	qInstallMsgHandler(myMessageOutput);
#endif
}

TextModPlugin::~TextModPlugin()
{
	delete ctxMenuCpp_;
	delete ctxMenuHtml_;
	delete menu_;
	delete toolBar_;
}

QString TextModPlugin::name() const
{
	return "TextMod";
}

QString TextModPlugin::description() const
{
	return "This plugin can insert different pre-defined text templates and convert existing text";
}

QString TextModPlugin::targetEngine() const
{
	return "sci";
}

Juff::ActionList TextModPlugin::contextMenuActions() const
{
	Juff::ActionList list;
	list << ctxMenuCpp_->menuAction()
		 << ctxMenuHtml_->menuAction();
	return list;
}

QMenu* TextModPlugin::menu() const
{
	return menu_;
}

QToolBar* TextModPlugin::toolBar() const
{
	return toolBar_;
}

QWidget* TextModPlugin::settingsPage() const
{
	return dlgCppSettings_;
}

void TextModPlugin::applySettings()
{
	if(0 == dlgCppSettings_)
		return;

	dlgCppSettings_->getSettings(sett_);
	sett_.save(this);
}

void TextModPlugin::toUpper()
{
	QString selText;
	manager()->getSelectedText(selText);
	if ( !selText.isEmpty() )
		manager()->replaceSelectedText(selText.toUpper());
}

void TextModPlugin::toLower()
{
	QString selText;
	manager()->getSelectedText(selText);
	if ( !selText.isEmpty() )
		manager()->replaceSelectedText(selText.toLower());
}

void TextModPlugin::date()
{
	DateFormatDlg dlg;
	
	if ( dlg.exec() == QDialog::Accepted )
	{
		QString fmt = dlg.format();
		QDateTime dt = QDateTime::currentDateTime();
		QString dateStr = dt.toString(fmt);
		int row, col;
		manager()->getCursorPos(row, col);
		manager()->insertText(dateStr);
		manager()->setCursorPos(row, col + dateStr.length());
	}
}

void TextModPlugin::cppNewClass()
{
	QString docName;
	manager()->getCurDocName(docName);
	QString subDirectory;
	QFileInfo fi(docName);
	if(0 == fi.suffix().compare(sett_.hppExt_, Qt::CaseInsensitive))
		subDirectory = sett_.hppSubdirectory_;
	else if(0 == fi.suffix().compare(sett_.cppExt_, Qt::CaseInsensitive))
		subDirectory = sett_.cppSubdirectory_;
	DlgClassProperties* dlg =
		new DlgClassProperties(docName,
							   subDirectory,
							   manager()->mainWindow());

	if(QDialog::Accepted != dlg->exec())
		return;

	// check paths, and create if they do not exist and the user agrees
	QString hppDir = dlg->path(),
			cppDir = dlg->path();
	if(!sett_.hppSubdirectory_.isEmpty())
		hppDir.append("\\").append(sett_.hppSubdirectory_);
	if(!sett_.cppSubdirectory_.isEmpty())
		cppDir.append("\\").append(sett_.cppSubdirectory_);
	QDir di(hppDir);
	if(!di.exists())
	{
		if(QMessageBox::Yes !=
				QMessageBox::question(manager()->mainWindow(), "",
									  QObject::tr("Directory \"%1\" does not exist - Create?")
											.arg(hppDir),
									  QMessageBox::Yes | QMessageBox::No,
									  QMessageBox::Yes))
			return;
		di.mkpath(hppDir);
	}
	if(cppDir != hppDir)
	{
		di.setPath(cppDir);
		if(!di.exists())
		{
			if(QMessageBox::Yes !=
					QMessageBox::question(manager()->mainWindow(), "",
										  QObject::tr("Directory \"%1\" does not exist - Create?")
												  .arg(cppDir),
										  QMessageBox::Yes | QMessageBox::No,
										  QMessageBox::Yes))
				return;
			di.mkpath(cppDir);
		}
	}

	// create header and source file names and if they exist, ask whether
	// to replace them or not
	enum { header = 0, source = 1 };
	QStringList files;
	QString fn = QString("%1\\%2.%3").arg(hppDir).arg(dlg->className()).arg(sett_.hppExt_);
	fn.replace(QString("..%1").arg(sett_.hppExt_), QString(".%1").arg(sett_.hppExt_));
	files.append(fn);
	fn = QString("%1\\%2.%3").arg(cppDir).arg(dlg->className()).arg(sett_.cppExt_);
	fn.replace(QString("..%1").arg(sett_.cppExt_), QString(".%1").arg(sett_.cppExt_));
	files.append(fn);
	for(QStringList::const_iterator it = files.begin(); it != files.end(); ++it)
	{
		QFileInfo fi(*it);
		if(fi.exists() &&
		   (QMessageBox::Yes != QMessageBox::question(
				manager()->mainWindow(), "",
				QObject::tr("File \"%1\" already exists - Replace?")
						.arg(*it),
				QMessageBox::Yes | QMessageBox::No,
				QMessageBox::Yes)))
			return;
	}

	// write the new header file and open it
	QFile hf(files.at(header));
	if(!hf.open(QIODevice::WriteOnly | QIODevice::Truncate))
	{
		QMessageBox::warning(manager()->mainWindow(), "",
							 QObject::tr("Could not create header file"));
		return;
	}
	QTextStream ht(&hf);
	QString txt;
	applyTemplate(txt, sett_.hppTemplate_, dlg->className(),
				  dlg->parentName(), dlg->accessSpec(), dlg->virtualDestructor());
	ht << txt;
	hf.close();
	manager()->openDoc(files.at(header));

	// write the new source file and open it
	QFile sf(files.at(source));
	if(!sf.open(QIODevice::WriteOnly | QIODevice::Truncate))
	{
		QMessageBox::warning(manager()->mainWindow(), "",
							 QObject::tr("Could not create source file"));
		return;
	}
	QTextStream st(&sf);
	applyTemplate(txt, sett_.cppTemplate_, dlg->className(),
				  dlg->parentName(), dlg->accessSpec(), dlg->virtualDestructor());
	st << txt;
	sf.close();
	manager()->openDoc(files.at(source));

	// switch to the header file
	manager()->openDoc(files.at(header));
}

void TextModPlugin::cppNewFunc()
{
	// get document text and split into "items" that define classes
	QString docText;
	manager()->getCurrentDocText(docText);
	QRegExp rexp("(\\b(class|struct)\\b|"
				 "\\{|"
				 "\\}\\s*;|"
				 "\\bpublic(|\\s+slots)\\s*:|"
				 "\\bprotected(|\\s+slots)\\s*:|"
				 "\\bprivate(|\\s+slots)\\s*:|"
				 "\\bsignals\\s*:|"
				 "//.*\\n|"
				 "/\\*.*\\*/)");
	rexp.setMinimal(true);
	ClassItem::Vec items;
	QString txt = docText,
			lastSep;
	for(int tinx = rexp.indexIn(txt); 0 <= tinx; tinx = rexp.indexIn(txt))
	{
		QString seg = txt.left(tinx),
				sep = rexp.cap(0);
		txt = txt.mid(tinx + sep.size());
		items.push_back(ClassItem(lastSep, seg));
		lastSep = sep;
	}
	items.push_back(ClassItem(lastSep, txt));

	// create a tree structure with potential class definitions
	ClassDesc::Vec classTree;
	QStringList incomplete;
	ClassDesc::Ptr cl;
	unsigned long i = 0;
	while(true)
	{
		cl = ClassDesc::create(i, items, "", incomplete);
		if(!cl)
			break;
		classTree.push_back(cl);
	}

	// flatten the tree structure
	ClassDesc::Vec classes;
	for(ClassDesc::ConstIt it = classTree.begin(); it != classTree.end(); ++it)
	{
		classes.append(*it);
		classes += (*it)->classVec();
	}

	// warning if no class declarations were found at all
	if(classes.isEmpty())
	{
		QMessageBox::warning(manager()->mainWindow(),
							 QObject::tr("Warning"),
							 QObject::tr("No class/struct declarations found:\n"
										 "Only global functions can be generated"));
	}

	// warning message if we are having incomplete classes
	else if(!incomplete.isEmpty())
	{
		QMessageBox::warning(manager()->mainWindow(),
							 QObject::tr("Warning"),
							 QObject::tr("Incomplete class/struct declarations found:\n- %1\n"
										 "Note that functions cannot be generated for them")
							 .arg(incomplete.join("\n- ")));
	}

	// list of available classes in this header file
	QStringList clist;
	for(ClassDesc::It it = classes.begin(); it != classes.end(); ++it)
		clist.append((*it)->declaration());
	clist.append(QObject::tr("(global function)"));

	// find out if the current file is possibly a source file
	// note that in this case it is assumed that the user wants the declaration
	// locally in this source file, and the definition in the same file, so no
	// further source file needs to be specified
	QString fName;
	manager()->getCurDocName(fName);
	QString srcName = fName;
	QString fExt = srcName.mid(srcName.lastIndexOf("."));
	bool currentIsSource = 0 == fExt.compare(sett_.cppExt_, Qt::CaseInsensitive);

	// propose a source file directory and name, based on C++ settings
	// and always assuming that the current file is a header file
	// if it is not, an empty file path will be passed to the dialog
	if(currentIsSource)
	{
		srcName = "";
	}
	else
	{
		QStringList fPath = srcName.split(QRegExp("[/\\\\]"), QString::SkipEmptyParts);
		QStringList incPath = sett_.hppSubdirectory_.split(QRegExp("[/\\\\]"), QString::SkipEmptyParts);
		int srcInx = fPath.size() - incPath.size() - 1;
		bool containsInc = true;
		int cpInx = srcInx;
		for(QStringList::const_iterator it = incPath.begin();
			(0 >= srcInx) && containsInc && (it != incPath.end()); ++it)
			containsInc = 0 == it->compare(fPath.at(cpInx++), Qt::CaseInsensitive);
		if(containsInc)
		{
			for(int i = 0; i < incPath.size(); ++i)
				fPath.removeAt(srcInx);
		}
		QStringList srcPath = sett_.cppSubdirectory_.split(QRegExp("[/\\\\]"), QString::SkipEmptyParts);
		if(!srcPath.isEmpty())
			fPath.insert(srcInx, srcPath.join("/"));
		srcName = fPath.join("/");
		int extInx = srcName.lastIndexOf('.');
		if(0 > extInx)
			extInx = srcName.size();
		srcName = srcName.left(extInx).append(".").append(sett_.cppExt_);
	}

	// display function properties dialog
	DlgFunctionProperties* dlg =
		new DlgFunctionProperties(clist, srcName,
								  manager()->mainWindow());
	if(QDialog::Accepted != dlg->exec())
		return;

	// from here only if the dialog was accepted
	int declSeg = -1;
	QString tabs(dlg->className().split("::").size(), QChar('\t'));

	// check if the definition goes also to the current file, even if it
	// is not already a source file
	if(!currentIsSource)
	{
		QFileInfo sourceFileInfo(dlg->sourceFile());

		if(!dlg->sourceFile().isEmpty() && !sourceFileInfo.exists())
			QMessageBox::information(manager()->mainWindow(), "",
									 QObject::tr("Note: File %1 does not exist")
									 .arg(dlg->sourceFile()));

		QString canonicalSourceFile = sourceFileInfo.canonicalFilePath();

		if(dlg->isInline() ||
		   dlg->sourceFile().isEmpty() ||
		   canonicalSourceFile.isEmpty() ||
		   (0 == canonicalSourceFile.compare(fName, Qt::CaseInsensitive)))
			currentIsSource = true;
	}

	// construction of declaration and definition strings
	QString decl = QString("%1%2%3%4 %5(%6)%7;\n\n")
				.arg(dlg->isStatic() ? "static " : "")
				.arg(dlg->isVirtual() ? "virtual " : "")
				.arg(dlg->isInline() ? "inline " : "")
				.arg(dlg->returnType())
				.arg(dlg->functionName())
				.arg(dlg->parameters())
				.arg(dlg->isConst() ? " const" : ""),
			def = QString("%1%2 %3::%4(%5)%6\n{\n}\n")
				.arg(dlg->isInline() ? "inline " : "")
				.arg(dlg->returnType())
				.arg(dlg->className())
				.arg(dlg->functionName())
				.arg(dlg->parametersWithoutDefaults())
				.arg(dlg->isConst() ? " const" : "");

	// declaration - global function
	if(dlg->isGlobal())
	{
		// find first class declaration
		// note that in case there is no class declaration at all, this will
		// point to one more than the last segment
		unsigned long firstClass = (unsigned long)items.size();
		for(ClassDesc::ConstIt it = classes.begin(); it != classes.end(); ++it)
			firstClass = qMin(firstClass, (*it)->nameInx_);

		// append the new declaration to the previous segment
		// note that even if there is a class declaration starting with the very
		// first letter, an empty segment 0 will still be before it, so the
		// minus 1 will never cause an underflow
		declSeg = firstClass - 1;
	}

	// declaration - class function:
	else
	{
		// find the class declaration
		// note: assure that there is never a 0 pointer, but should not happen anyway!
		ClassDesc::Ptr cl = *classes.begin();
		for(ClassDesc::ConstIt it = classes.begin(); it != classes.end(); ++it)
		{
			if((*it)->name_ == dlg->className())
			{
				cl = *it;
				break;
			}
		}

		// see if a segment of desired type already exists
		ClassDesc::InxVec* segVec = 0;
		QString accessSpec = dlg->accessSpec();
		if("public" == accessSpec)
			segVec = &cl->publicSegs_;
		else if("protected" == accessSpec)
			segVec = &cl->protectedSegs_;
		else if("private" == accessSpec)
			segVec = &cl->privateSegs_;
		else if("public slots" == accessSpec)
			segVec = &cl->publicSlotsSegs_;
		else if("protected slots" == accessSpec)
			segVec = &cl->protectedSlotsSegs_;
		else if("private slots" == accessSpec)
			segVec = &cl->privateSlotsSegs_;
		else if("signals" == accessSpec)
			segVec = &cl->signalsSegs_;

		// if a segment already exists, we use the last occurrence
		if(!segVec->isEmpty())
		{
			declSeg = segVec->back();
			items[declSeg].second.append(tabs);
		}

		// if no segment exists, we prepend one to the closing bracket
		else
		{
			declSeg = cl->closeInx_ - 1;
			items[declSeg].second.append(accessSpec).append(":\n").append(tabs);
		}
	}

	// insert the declaration and put text back together
	items[declSeg].second.append(decl);
	QString newDocText;
	int newRow, newCol;
	for(int i = 0; i < items.size(); ++i)
	{
		newDocText.append(items[i].key()).append(items[i].data());
		if(i == (int)declSeg)
		{
			rowColEndText(newDocText, newRow, newCol);
			newRow -= 2;
			newCol = 0;
		}
	}

	// if the current is not the source, save text and go to source
	if(!currentIsSource)
	{
		// replace current text
		int rows, cols;
		rowColEndText(docText, rows, cols);
		manager()->setSelection(0, 0, rows, cols);
		manager()->removeSelectedText();
		manager()->insertText(newDocText);

		// set cursor to new class
		manager()->setCursorPos(newRow, newCol);

		// go to source file
		manager()->openDoc(dlg->sourceFile());
		manager()->getCurrentDocText(docText);
		newDocText = docText;
	}

	// avoid putting definition outside any ifndef/endif bracket, so
	// we need to split the text again
	QRegExp rexp2("#ifndef|"
				  "#endif|"
				  "(\\b(class|struct)\\b|"
				  "//.*\\n|"
				  "/\\*.*\\*/)");
	rexp2.setMinimal(true);
	items.clear();
	txt = newDocText;
	lastSep = "";
	for(int tinx = rexp2.indexIn(txt); 0 <= tinx; tinx = rexp2.indexIn(txt))
	{
		QString seg = txt.left(tinx),
				sep = rexp2.cap(0);
		txt = txt.mid(tinx + sep.size());
		items.push_back(ClassItem(lastSep, seg));
		lastSep = sep;
	}
	items.push_back(ClassItem(lastSep, txt));

	// see if there is a starting ifndef with only empty or comment segs before
	int defSeg = -1;
	for(defSeg = 0; defSeg < items.size(); ++defSeg)
	{
		QString key = items.at(defSeg).key();

		// if the segment is a class/struct definition, we assume that there is
		// no ifndef/endif bracketing the file
		if((key == "class") || (key == "struct"))
		{
			defSeg = items.size();
			break;
		}

		// if a ifndef appears, we are ok
		if(key == "#ifndef")
			break;

		// if something that is neither empty or a comment appears, we also assume
		// that there is no ifndef/endif
		if(!key.isEmpty() &&
		   !key.startsWith('/'))
		{
			defSeg = items.size();
			break;
		}
	}

	// if we found a starting ifndef, we search for the corresponding endif
	// from the end
	int endIf = 0;
	if(defSeg < items.size())
	{
		for(endIf = items.size() - 1; endIf > 0; --endIf)
		{
			QString key = items.at(defSeg).key();

			// check for an endif
			if(key == "#endif")
				break;

			// otherwise we only accept empty or comment items
			if(!key.isEmpty() &&
			   !key.startsWith('/'))
			{
				endIf = 0;
				break;
			}
		}
	}

	// now find the correct item to append the definition
	if(defSeg < endIf)
		defSeg = endIf - 1;
	else
		defSeg = items.size() - 1;

	// insert the definition and put text back together
	items[defSeg].second.append(def);
	docText = newDocText;
	newDocText = "";
	for(int i = 0; i < items.size(); ++i)
	{
		newDocText.append(items[i].key()).append(items[i].data());
		if(i == defSeg)
		{
			rowColEndText(newDocText, newRow, newCol);
			newRow -= 2;
			newCol = 0;
		}
	}

	// replace current text
	int rows, cols;
	rowColEndText(docText, rows, cols);
	manager()->setSelection(0, 0, rows, cols);
	manager()->removeSelectedText();
	manager()->insertText(newDocText);

	// set cursor to new class
	manager()->setCursorPos(newRow, newCol);
}

void TextModPlugin::htmlTemplate()
{
	QString text = "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n"
			"<html>\n"
			"<head>\n"
			"<title></title>\n"
			"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" >\n"
			"<link type=\"text/css\" rel=\"stylesheet\" href=\"style.css\">\n"
			"</head>\n"
			"<body>\n\n"
			"</body>\n"
			"</html>\n";
	int row, col;
	manager()->getCursorPos(row, col);
	manager()->insertText(text);
	manager()->setCursorPos(row + 3, 7);
}

void TextModPlugin::htmlTable()
{
	QString text = "<table>\n"
			"  <tr>\n"
			"    <td></td> <td></td>\n"
			"  </tr>\n"
			"  <tr>\n"
			"    <td></td> <td></td>\n"
			"  </tr>\n"
			"</table>\n";
	int row, col;
	manager()->getCursorPos(row, col);
	manager()->insertText(text);
	manager()->setCursorPos(row + 2, 8);
}

void TextModPlugin::htmlHref()
{
	wrapSelectedText("<a href=\"\">%1</a>", 9);
}

void TextModPlugin::htmlImg()
{
	QString text = "<img src=\"\" alt=\"\" />";
	int row, col;
	manager()->getCursorPos(row, col);
	manager()->insertText(text);
	manager()->setCursorPos(row, col + 10);
}

void TextModPlugin::htmlBold()
{
	wrapSelectedText("<b>%1</b>", 3);
}

void TextModPlugin::htmlItalic()
{
	wrapSelectedText("<i>%1</i>", 3);
}

void TextModPlugin::htmlUnderline()
{
	wrapSelectedText("<u>%1</u>", 3);
}

void TextModPlugin::htmlDiv()
{
	wrapSelectedText("<div>%1</div>", 5);
}

void TextModPlugin::htmlUList()
{
	createList(false);
}

void TextModPlugin::htmlOList()
{
	createList(true);
}

void TextModPlugin::callColorDlg()
{
	QColor color = QColorDialog::getColor();
	if ( color.isValid() )
		manager()->insertText(color.name());
}

void TextModPlugin::createList(bool numerated)
{
	QString tag = numerated ? "ol" : "ul";
	QString selText;
	manager()->getSelectedText(selText);

	if ( !selText.isEmpty() )
	{
		//	set selection and split it into lines
		int row1, col1, row2, col2;
		manager()->getSelection(row1, col1, row2, col2);
		QStringList lines = selText.split(QRegExp("(\n|\r|\r\n)"));
		if ( col2 == 0 )	//	last string is empty
			lines.pop_back();
		
		//	create a list
		QString text = QString("<%1>\n").arg(tag);
		if ( col1 > 0 )		//	go to the next line
			text.prepend("\n");
		foreach (QString line, lines)
			text = text + "  <li>" + line + "\n";
		text += QString("</%1>").arg(tag);

		manager()->replaceSelectedText(text);
	}
	else
	{
		int row, col;
		manager()->getCursorPos(row, col);
		manager()->insertText(QString("<%1>\n  <li>\n  <li>\n  <li>\n</%2>").arg(tag).arg(tag));
		manager()->setCursorPos(row + 1, 6);
	}
}

void TextModPlugin::wrapSelectedText(const QString& fmt, int indent)
{
	QString selText;
	manager()->getSelectedText(selText);
	if ( !selText.isEmpty() )
	{
		int row1, col1, row2, col2;
		manager()->getSelection(row1, col1, row2, col2);
		manager()->replaceSelectedText(fmt.arg(selText));
		manager()->setCursorPos(row1, col1 + indent);
	}
	else
	{
		int row, col;
		manager()->getCursorPos(row, col);
		manager()->insertText(fmt.arg(""));
		manager()->setCursorPos(row, col + indent);
	}
}

void TextModPlugin::applyTemplate(QString& text, QString const& tpl,
								  QString const& className,
								  QString const& parentName,
								  QString const& accessSpec,
								  bool virtualDestructor)
{
	QDateTime dt = QDateTime::currentDateTime();
	QString dateStr = dt.toString(Qt::ISODate);
	QString includeParent,
			parentAccess,
			constructParent;
	if(!parentName.isEmpty())
	{
		includeParent = sett_.includeParent_;
		parentAccess = sett_.parentAccess_;
		parentAccess.replace("[qualifier]", accessSpec, Qt::CaseInsensitive);
		constructParent = sett_.constructParent_;
	}
	text = tpl;
	text.replace("[include_parent]", includeParent, Qt::CaseInsensitive)
		.replace("[parent_access]", parentAccess, Qt::CaseInsensitive)
		.replace("[construct_parent]", constructParent, Qt::CaseInsensitive)
		.replace("[CLASS]", className.toUpper())
		.replace("[class]", className, Qt::CaseInsensitive)
		.replace("[HPP_EXT]", sett_.hppExt_.toUpper())
		.replace("[hpp_ext]", sett_.hppExt_, Qt::CaseInsensitive)
		.replace("[CPP_EXT]", sett_.cppExt_.toUpper())
		.replace("[cpp_ext]", sett_.cppExt_, Qt::CaseInsensitive)
		.replace("[parent]", parentName, Qt::CaseInsensitive)
		.replace("[date]", dateStr, Qt::CaseInsensitive);
	if(virtualDestructor)
	{
		// this replaces only the declaration (with ;), not the function definition,
		// and still allows for whitespace everywhere - for people who like that
		text.replace(QRegExp(QString("~%1(\\s)*\\((\\s)*\\)(\\s)*;").arg(className)),
					 QString("virtual ~%1\\1(\\2)\\3;").arg(className));
	}
}

void TextModPlugin::rowColEndText(QString const& txt, int& row, int& col)
{
	QStringList txtList = txt.split("\n");
	row = txtList.size() - 1;
	col = txtList.back().size() - 1;
}

Q_EXPORT_PLUGIN2(textmod, TextModPlugin)
